const ALL_SOUND_OFF_EVENT_DATA = Array.from({ length: 16 }, (_, index) => new Uint8Array([176 + index, 120, 0]));
export class MidiPlayer {
    constructor({ encodeMidiMessage, filterMidiMessage, json, midiFileSlicer, midiOutput, startScheduler }) {
        this._encodeMidiMessage = encodeMidiMessage;
        this._filterMidiMessage = filterMidiMessage;
        this._json = json;
        this._midiFileSlicer = midiFileSlicer;
        this._midiOutput = midiOutput;
        this._startScheduler = startScheduler;
        this._state = null;
    }
    get state() {
        return this._state === null ? 'stopped' : this._state.stopScheduler === null ? 'paused' : 'playing';
    }
    pause() {
        if (this._state === null || this._state.stopScheduler === null) {
            throw new Error('The player is not playing.');
        }
        this._clear();
        const { resolve, stopScheduler } = this._state;
        this._state.offset = stopScheduler();
        this._state.stopScheduler = null;
        resolve();
    }
    play() {
        if (this._state !== null) {
            throw new Error('The player is not stopped.');
        }
        return this._schedule(0, 0);
    }
    resume() {
        if (this._state === null || this._state.stopScheduler !== null) {
            throw new Error('The player is not paused.');
        }
        const { endedTracks, offset } = this._state;
        this._state = null;
        return this._schedule(endedTracks, offset);
    }
    stop() {
        if (this._state === null) {
            throw new Error('The player is already stopped.');
        }
        if (this._state.stopScheduler !== null) {
            this._clear();
            this._stop(this._state);
        }
    }
    _clear() {
        var _a, _b;
        // Bug #1: Chrome does not yet implement the clear() method.
        (_b = (_a = this._midiOutput).clear) === null || _b === void 0 ? void 0 : _b.call(_a);
        ALL_SOUND_OFF_EVENT_DATA.forEach((data) => this._midiOutput.send(data));
    }
    _schedule(endedTracks, offset) {
        return new Promise((resolve) => {
            const stopScheduler = this._startScheduler(({ end, start }) => {
                if (this._state === null) {
                    this._state = { endedTracks, offset: start - offset, resolve, stopScheduler: null };
                }
                const events = this._midiFileSlicer.slice(start - this._state.offset, end - this._state.offset);
                events
                    .filter(({ event }) => this._filterMidiMessage(event))
                    .forEach(({ event, time }) => this._midiOutput.send(this._encodeMidiMessage(event), start + time));
                this._state.endedTracks += events.filter(({ event }) => MidiPlayer._isEndOfTrack(event)).length;
                if (this._state.endedTracks === this._json.tracks.length) {
                    this._stop(this._state);
                }
            });
            if (this._state === null) {
                stopScheduler();
            }
            else {
                this._state.stopScheduler = stopScheduler;
            }
        });
    }
    _stop(state) {
        const { resolve, stopScheduler } = state;
        stopScheduler === null || stopScheduler === void 0 ? void 0 : stopScheduler();
        this._state = null;
        resolve();
    }
    static _isEndOfTrack(event) {
        return 'endOfTrack' in event;
    }
}
//# sourceMappingURL=midi-player.js.map